--[[
	Copyright (C) 2019 noobfighting < lukaskoschine@gmail.com >

	This file is part of PHIL.
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
]]

PHIL = LibStub("AceAddon-3.0"):NewAddon("PHIL", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("PHIL")

PHIL_SIMPLE_NAME = L["PHIL_PRINT"]
PHIL_VERSION = "v1.0.7"
PHIL_PRINT = PHIL_SIMPLE_NAME .. " " .. PHIL_VERSION

local PHILSimpleName = PHIL_SIMPLE_NAME
local PHILPrint = PHIL_PRINT

local db

local frame = CreateFrame("FRAME")

--ADDON EVENTS--
local EVENT_ADDON_LOADED = "ADDON_LOADED"
local EVENT_PLAYER_LOGOUT = "PLAYER_LOGOUT"
local EVENT_INSPECT_READY = "INSPECT_READY"
local EVENT_UNIT_INVENTORY_CHANGED = "UNIT_INVENTORY_CHANGED"

local unitsToReinspect = {}
local unitsToInspect = {}
local inspectedUnits = {}

local REINSPECTION_INTERVAL = 20
local INSPECTION_INTERVAL = 1

local lastInspectionTime = time()

local iLvlCache = {}
local Slots = {"HeadSlot", "NeckSlot", "ShoulderSlot", "BackSlot", "ChestSlot", "WristSlot", "HandsSlot", "WaistSlot", "LegsSlot", "FeetSlot", "Finger0Slot", "Finger1Slot", "Trinket0Slot", "Trinket1Slot", "MainHandSlot", "SecondaryHandSlot"}
local twoHandedSubTypes = {LE_ITEM_WEAPON_AXE2H, LE_ITEM_WEAPON_STAFF, LE_ITEM_WEAPON_BOWS, LE_ITEM_WEAPON_GUNS, LE_ITEM_WEAPON_MACE2H, LE_ITEM_WEAPON_SWORD2H, LE_ITEM_WEAPON_STAFF, LE_ITEM_WEAPON_CROSSBOW, LE_ITEM_WEAPON_FISHINGPOLE, LE_ITEM_WEAPON_POLEARM}
local tooltip = CreateFrame("GameTooltip", "iLevelScanTooltip", nil, "GameTooltipTemplate")
tooltip:SetOwner(UIParent, "ANCHOR_NONE")


function PHIL:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New(PHIL_SIMPLE_NAME, defaults, true)
    db = self.db.profile
    self:Print(PHILPrint)
end

function frame:OnEvent(event, arg1, arg2, arg3)
    if event == EVENT_ADDON_LOADED and arg1 == PHIL_SIMPLE_NAME then
	elseif event == EVENT_INSPECT_READY then
		local name = GetActualUnitName(arg1)
		local oldItemLevel = iLvlCache[name]
		UpdateAverageItemLevel(name)
		ShowAverageItemLevel(name)
		if iLvlCache[name] and name == GameTooltip:GetUnit() and oldItemLevel ~= iLvlCache[name] then
			UpdateTooltip(GameTooltip, iLvlCache[name], name)
		end
	elseif event == EVENT_UNIT_INVENTORY_CHANGED then
		local name = GetActualUnitName(arg1)
		local oldItemLevel = iLvlCache[name]
		UpdateAverageItemLevel(name)
		if iLvlCache[name] and name == GameTooltip:GetUnit() and oldItemLevel ~= iLvlCache[name] then
			UpdateTooltip(GameTooltip, iLvlCache[name], name)
		end
    end
end

function GetActualUnitName(rawName)
	local name
	if rawName then
		if rawName == "player" then
			name = GetUnitName(rawName)
		else
			_, _, _, _, _, name = GetPlayerInfoByGUID(rawName)
		end
	end
	return name
end

frame:RegisterEvent(EVENT_ADDON_LOADED)
frame:RegisterEvent(EVENT_PLAYER_LOGOUT)
frame:RegisterEvent(EVENT_INSPECT_READY)
frame:RegisterEvent(EVENT_UNIT_INVENTORY_CHANGED)

frame:SetScript("OnEvent", frame.OnEvent)
frame:SetScript("OnUpdate", function() ProcessInspection(); end)

function UpdateTooltip(tooltip, newItemLevel, unit)
	local lines = {}
	for i = 1, tooltip:NumLines() do
		local left = _G[tooltip:GetName() .. "TextLeft" .. i]
		local lText = left:GetText()
		local lR, lG, lB = left:GetTextColor()
		if lText:find(L["ITEM_LEVEL_PRINT"] .. "(%d+)") then
			tinsert(lines, {line = L["ITEM_LEVEL_PRINT"] .. newItemLevel, r = lR, g = lG, b = lB})
		else
			tinsert(lines, {line = lText, r = lR, g = lG, b = lB})
		end
	end
	tooltip:ClearLines()
	for k, v in pairs(lines) do
		tooltip:AddLine(v.line)
		_G[tooltip:GetName() .. "TextLeft" .. k]:SetTextColor(v.r, v.g, v.b)
	end
	tooltip:SetUnit(unit)
end

function OnShowGameTooltip(arg1, arg2)
	local unitName = GameTooltip:GetUnit()
	if CheckIfInspectable(unitName) then
		local unit = getByNameAttribute(inspectedUnits, unitName)
		if unit then
			if not getByNameAttribute(unitsToReinspect, unit.name) then
				table.insert(unitsToReinspect, unit)
			end
		else
			if not getByNameAttribute(unitsToInspect, unitName) then
				local unitToInspect = {}
				unitToInspect.name = unitName
				table.insert(unitsToInspect, unitToInspect)
				--double inspect on first inspection due to item loading issue
				table.insert(unitsToInspect, unitToInspect)
			end
		end
	end
	ShowAverageItemLevel(arg1, arg2)
end

function ProcessInspection()
	local currentTime = time()
	if (currentTime - lastInspectionTime) > INSPECTION_INTERVAL then
		local inspectableUnit = GetFirstInspectableUnit(unitsToInspect)
		if inspectableUnit then
			lastInspectionTime = currentTime
			removeByName(unitsToInspect, inspectableUnit.name)
			local inspectedUnit = getByNameAttribute(inspectedUnits, unitName)
			if inspectedUnit then
				inspectedUnit.last = currentTime
				ClearInspectPlayer(inspectedUnit.name)
			else
				inspectableUnit.last = currentTime
				table.insert(inspectedUnits, inspectableUnit)
			end
			NotifyInspect(inspectableUnit.name)
		else
			local reinspectableUnit = GetFirstInspectableUnit(unitsToReinspect, currentTime)
			if reinspectableUnit then
				lastInspectionTime = currentTime
				removeByName(unitsToReinspect, reinspectableUnit.name)
				reinspectableUnit.last = currentTime
				ClearInspectPlayer(reinspectableUnit.name)
				NotifyInspect(reinspectableUnit.name)
			end
		end
	end
end

function removeByName(array, name)
	for i=#array,1,-1 do
		if array[i].name == name then
			table.remove(array, i)
			break
		end
	end
end

function UnitInThisPartyOrRaid(unit)
	return UnitIsUnit(unit, "player") or UnitInParty(unit) or UnitInRaid(unit)
end

function ShowAverageItemLevel(arg1, arg2)
	local Unit = GameTooltip:GetUnit()
	if Unit and not tooltipContainsILvl(GameTooltip:GetRegions()) then
		if iLvlCache[Unit] then
			GameTooltip:AddLine(L["ITEM_LEVEL_PRINT"] .. iLvlCache[Unit])
			GameTooltip:Show()
		end
	end
end

function UpdateAverageItemLevel(unitName)
	if CheckIfInspectable(unitName) then 
		averageItemLevel = CalculateAverageItemLevel(unitName)
		if averageItemLevel ~= 0 then
			iLvlCache[unitName] = averageItemLevel
		end
	end
end

function GetItemLevelForItemLink(itemLink)
	if not itemLink then return end
 
	tooltip:ClearLines()
	tooltip:SetHyperlink(itemLink)

	for i = 2, tooltip:NumLines() do
		local text = _G["iLevelScanTooltipTextLeft" .. i]:GetText()
		if(text and text ~= "") then
			_, _, ilvl = string.find(text, L["ITEM_LEVEL_SEARCH"] .. "%s(%d+)")
			if ilvl and ilvl ~= "" then
				return tonumber(ilvl)
			end
		end
	end
	return 0
end

function CalculateAverageItemLevel(Unit)
	local iLvlSum = 0
	for k, Slot in pairs(Slots) do
		local slotId = GetInventorySlotInfo(Slot)
		if slotId then
			GetInventoryItemLink(Unit, slotId)
			local iLink = GetInventoryItemLink(Unit, slotId)
			if iLink then
				local iLvl = GetItemLevelForItemLink(iLink)
				if iLvl ~= 0 then
					--for weighting purposes single 2H weapons count twice, as main AND offhand slot
					--specs that can wear two 2H weapons are not considered
					if Slot == "MainHandSlot" then
						local _, _, _, _, _, _, classID, subclassID = select(6, GetItemInfo(iLink))
						if contains(twoHandedSubTypes, subclassID) then
							local offHandSlotId = GetInventorySlotInfo("SecondaryHandSlot")
							if not offHandSlotId or not GetInventoryItemLink(Unit, offHandSlotId) then
								iLvlSum = iLvlSum + iLvl
							end
						end
					end
					iLvlSum = iLvlSum + iLvl
				end
			end
		end
	end
	return math.floor(iLvlSum / 16)
end

function CheckIfInspectable(unitName)
	return unitName and UnitInThisPartyOrRaid(unitName) and CanInspect(unitName) and CheckInteractDistance(unitName, 1)
end

function tooltipContainsILvl(...)
    for i = 1, select("#", ...) do
        local region = select(i, ...)
        if region and region:GetObjectType() == "FontString" then
            local text = region:GetText()
			if text ~= nil and string.find(text, L["ITEM_LEVEL_PRINT"] .. "(%d+)") then
				return true
			end
        end
    end
	return false
end

function contains(keyWords, msg)
    for key, filter in ipairs(keyWords) do
        if filter == msg then
            return true
        end
    end
    return false
end

function getByNameAttribute(inspectedUnits, name)
    for key, unit in ipairs(inspectedUnits) do
        if unit.name == name then
            return unit
        end
    end
    return nil
end

function GetFirstInspectableUnit(unitsToInspect, currentTime)
    for key, unit in ipairs(unitsToInspect) do
        if CheckIfInspectable(unit.name) and (not currentTime or (currentTime - unit.last) > REINSPECTION_INTERVAL) then
            return unit
        end
    end
    return nil
end

GameTooltip:HookScript("OnTooltipSetUnit", OnShowGameTooltip)
GameTooltip:HookScript("OnShow", OnShowGameTooltip)
GameTooltip:HookScript("OnUpdate", ShowAverageItemLevel)